import json
import datetime
import pprint
import psycopg2
import tornado.web
import tornado.gen
from tornado.log import gen_log, app_log
import dbconnector.models as models


class DbConnectorHandler(tornado.web.RequestHandler):

    def __init__(self, *args, **kwargs):
        super(DbConnectorHandler, self).__init__(*args, **kwargs)

    def _log_error(self, desc, response):
        app_log.error("%(service_name)s:%(uri)s:%(desc)s:%(msg)s", {
            "service_name": self.service_name,
            "uri": self.request.uri,
            "desc": desc,
            "msg": pprint.pformat(response)
        })
        self.write(response)


    def _write_result(self, response):
        gen_log.debug("%(service_name)s:%(uri)s:Sending Response %(msg)s", {
            "service_name": self.service_name,
            "uri": self.request.uri,
            "msg": pprint.pformat(response),
        })
        self.write(response)


class DoesPasteExistController(DbConnectorHandler):

    def __init__(self, *args, **kwargs):
        self.service_name = "does paste exist"
        super(DoesPasteExistController, self).__init__(*args, **kwargs)

    @tornado.gen.coroutine
    def post(self):
        response = {"error": True, "msg": None}

        try:
            text_raw = self.get_body_argument('text')
            source = self.get_body_argument('source')
            query = json.loads(text_raw)
            query_key = query['key']
        except ValueError, exc:
            # Raised in the case of invalid JSON
            response['msg'] = str(exc)
        except tornado.web.MissingArgumentError, exc:
            response['msg'] = "Missing POST argument '{}'".format(exc.arg_name)
        except KeyError:
            response['msg'] = "Missing key and / or body value in query JSON"

        if response["msg"]:
            self._log_error("error parsing submission", response)
            return

        paste_id_future = models.paste_id_for_source_and_key(self.application.db,
                                                             source, query_key)
        try:
            cursor = yield paste_id_future
            results = cursor.fetchall()
        except (psycopg2.Warning, psycopg2.Error) as exc:
            app_log.error("%(service_name)s:Database Error: %(msg)s", {
                "service_name": self.service_name,
                "msg": str(exc)
            })
            self.write({"error": True, "msg": str(exc)})
            return

        try:
            paste_id = results[0][0]
            msg = "Found text with paste_id {}".format(paste_id)
        except IndexError, _:
            paste_id = None
            msg = "No matching text was found"

        response = {
            u"error": False,
            u"paste_id": paste_id,
            "msg": msg,
        }
        self._write_result(response)

def filter_input(a_dict, a_key):
    if a_key not in a_dict:
        return False

    val = a_dict[a_key]

    if isinstance(val, basestring):
        return len(val) > 0
    return True


_REQUIRED_FIELDS = ("key", "url", "text")
_OPTIONAL_FIELDS = ("syntax", "expires_on", "added_on", "user", "title", "scrape_url")
class RecordPasteController(DbConnectorHandler):

    def __init__(self, *args, **kwargs):
        self.service_name = "record paste"
        super(RecordPasteController, self).__init__(*args, **kwargs)


    @tornado.gen.coroutine
    def post(self):
        response = {"error": True, "msg": None}
        try:
            text_raw = self.get_body_argument('text')
            source = self.get_body_argument('source')
            is_dox = self.get_body_argument('is_dox') == "True"
            query = json.loads(text_raw)
        except ValueError, exc:
            # ValueError raised in the case of invalid JSON
            response['msg'] = str(exc)
        except tornado.web.MissingArgumentError, exc:
            response['msg'] = "Missing POST argument '{}'".format(exc.arg_name)

        if response["msg"]:
            self._log_error("error parsing submission", response)
            return

        try:
            required_values = [query[key] for key in _REQUIRED_FIELDS]
        except KeyError, exc:
            response['msg'] = "Submitted text is missing a {} field".format(exc.message)
            self._log_error("missing key in dox json text", response)
            return

        optional_values = {k: query[k] for k in _OPTIONAL_FIELDS if filter_input(query, k)}

        for date_key in ('expires_on', 'added_on'):
            if date_key in optional_values:
                optional_values[date_key] = datetime.datetime.fromtimestamp(float(optional_values[date_key]))

        paste_id_future = models.insert_paste(self.application.db, source, is_dox,
                                              *required_values, **optional_values)
        try:
            cursor = yield paste_id_future
            results = cursor.fetchall()
        except (psycopg2.Warning, psycopg2.Error) as exc:
            app_log.error("%(service_name)s:Database Error: %(msg)s", {
                "service_name": self.service_name,
                "msg": str(exc)
            })
            self.write({"error": True, "msg": str(exc)})
            return

        try:
            paste_id = results[0][0]
            msg = "recorded text with paste_id {}".format(paste_id)
        except IndexError, _:
            paste_id = None
            msg = "-"

        response = {}
        response['error'] = False
        response['paste_id'] = paste_id
        response['msg'] = msg
        self._write_result(response)


class RecordSocialAccountController(DbConnectorHandler):

    def __init__(self, *args, **kwargs):
        self.service_name = "record social account"
        super(RecordSocialAccountController, self).__init__(*args, **kwargs)


    @tornado.gen.coroutine
    def post(self):
        response = {"error": True, "msg": None}
        try:
            paste_id = self.get_body_argument('paste_id')
            network = self.get_body_argument('service')
            account = self.get_body_argument('account')
        except tornado.web.MissingArgumentError, exc:
            response['msg'] = "Missing POST argument '{}'".format(exc.arg_name)

        if response['msg']:
            self._log_error("error parsing submission", response)
            return

        account_id_future = models.add_account_for_paste(self.application.db,
                                                         paste_id, network,
                                                         account)
        try:
            cursor = yield account_id_future
            results = cursor.fetchall()
        except (psycopg2.Warning, psycopg2.Error) as exc:
            app_log.error("%(service_name)s:Database Error: %(msg)s", {
                "service_name": self.service_name,
                "msg": str(exc)
            })
            self.write({"error": True, "msg": str(exc)})
            return

        response = {"msg": None, "error": True}
        try:
            account_id = results[0][0]
            response['msg'] = "Saved account with {}".format(account_id)
        except IndexError, exc:
            response['msg'] = str(exc)
            self._log_error("unexpected account id query results", response)
            return

        response['error'] = False
        response['account_id'] = account_id
        self._write_result(response)


class RecordSocialAccountCrawlController(DbConnectorHandler):


    def __init__(self, *args, **kwargs):
        self.service_name = "record social account crawl"
        super(RecordSocialAccountCrawlController, self).__init__(*args, **kwargs)


    @tornado.gen.coroutine
    def post(self):
        response = {"error": True, "msg": None}

        # Twitter account crawls will have a "mentions" key in the posted
        # JSON string.  No other crawled social accounts will.
        mentions = None

        try:
            account_id = self.get_body_argument('account_id')
            crawl_json = self.get_body_argument('crawl')
            crawl = json.loads(crawl_json)
            status = crawl['status']
            body = crawl['body']
            posts = crawl['posts']
            if "mentions" in crawl and crawl["mentions"]:
                mentions = crawl["mentions"]

        except ValueError, exc:
            # Raised in the case of invalid JSON
            response['msg'] = str(exc)
        except tornado.web.MissingArgumentError, exc:
            response['msg'] = "Missing POST argument '{}'".format(exc.arg_name)
        except KeyError, exc:
            # Raised when the submitted crawl JSON text doesn't contain all the
            # expected values
            response['msg'] = "Missing expected value in crawl JSON text: {}".format(exc.message)

        if response['msg']:
            self._log_error("error parsing submission", response)
            return


        crawl_id_future = models.add_crawl_for_account(self.application.db,
                                                       account_id, status, body)
        try:
            cursor = yield crawl_id_future
            results = cursor.fetchall()
        except (psycopg2.Warning, psycopg2.Error) as exc:
            app_log.error("%(service_name)s:Database Error: %(msg)s", {
                "service_name": self.service_name,
                "msg": str(exc)
            })
            self.write({"error": True, "msg": str(exc)})
            return

        response = {"error": True, "msg": None}

        try:
            crawl_id = results[0][0]
        except IndexError, exc:
            response['msg'] = str(exc)
            self._log_error("unexpected crawl id query results", response)
            return

        if mentions:
            try:
                mentions_ids_future = models.record_mentions_for_crawl(
                    self.application.db, crawl_id, mentions
                )
                cursor = yield mentions_ids_future
                results = cursor.fetchall()
            except (psycopg2.Warning, psycopg2.Error) as exc:
                app_log.error("%(service_name)s:Database Error: %(msg)s", {
                    "service_name": self.service_name,
                    "msg": str(exc)
                })
                self.write({"error": True, "msg": str(exc)})
                return
            except KeyError, exc:
                response['msg'] = "Missing a {} field".format(exc.message)
                self._log_error("missing key in mentions json text", response)
                return

        post_ids = []

        if posts is None or len(posts) == 0:
            response['error'] = False
            response['crawl_id'] = crawl_id
            response['post_ids'] = post_ids
            response['msg'] = "Successfully saved crawl (with no posts) with id {}".format(crawl_id)
            self._write_result(response)
            return

        post_ids_future = models.add_posts_to_crawl(self.application.db,
                                                    crawl_id, posts)
        try:
            cursor = yield post_ids_future
            results = cursor.fetchall()
        except (psycopg2.Warning, psycopg2.Error) as exc:
            app_log.error("%(service_name)s:Database Error: %(msg)s", {
                "service_name": self.service_name,
                "msg": str(exc)
            })
            self.write({"error": True, "msg": str(exc)})
            return

        response = {"error": True, "msg": None}

        try:
            post_ids = [row[0] for row in results]
            response['msg'] = "Successfully saved crawl (and {} posts) with id {}".format(len(post_ids), crawl_id)
        except IndexError, exc:
            response['msg'] = str(exc)
            self._log_error("unexpected post ids query results", response)
            return

        response['error'] = False
        response['crawl_id'] = crawl_id
        response['post_ids'] = post_ids
        self._write_result(response)


class AccountsForNthRecrawl(DbConnectorHandler):

    def __init__(self, *args, **kwargs):
        self.service_name = "accounts to recrawl"
        super(AccountsForNthRecrawl, self).__init__(*args, **kwargs)


    @tornado.gen.coroutine
    def post(self):
        response = {"error": True, "msg": None}
        try:
            recrawl_num = self.get_body_argument('num')
        except tornado.web.missingargumenterror, exc:
            response['msg'] = "missing post argument '{}'".format(exc.arg_name)
            self._log_error("error parsing submission", response)
            return

        try:
            recrawl_int = int(recrawl_num)
        except ValueError, exc:
            response['msg'] = ("Invalid 'int' argument for 'num', "
                               "received value '{}'".format(recrawl_num))
            self._log_error("error parsing submission", response)
            return

        if recrawl_int < 1 or recrawl_int > 10:
            response['msg'] = ("Invalid int recieved for recrawl number, must "
                               "be between 1 and 4, inclusive. "
                               "Received '{}'".format(recrawl_int))
            self._log_error("error parsing submission", response)
            return

        if recrawl_int == 1:
            date_offset = datetime.timedelta(1)
        elif recrawl_int == 2:
            date_offset = datetime.timedelta(2)
        elif recrawl_int == 3:
            date_offset = datetime.timedelta(3)
        elif recrawl_int == 4:
            date_offset = datetime.timedelta(7)
        else:
            weeks_back = (recrawl_int - 4 + 1)
            date_offset = datetime.timedelta(7 * weeks_back)

        start_datetime = datetime.datetime.now() - date_offset

        accounts_future = models.accounts_to_recrawl(self.application.db,
                                                     start_datetime, recrawl_int)

        try:
            cursor = yield accounts_future
            results = cursor.fetchall()
        except (psycopg2.Warning, psycopg2.Error) as exc:
            app_log.error("%(service_name)s:Database Error: %(msg)s", {
                "service_name": self.service_name,
                "msg": str(exc)
            })
            self.write({"error": True, "msg": str(exc)})
            return

        response = {"error": True, "msg": None}

        try:
            to_recrawl = [{"account_id": r[0], "account_name": r[1], "service_name": r[2]} for r in results]
            response['msg'] = "Successfully located {} accounts to recrawl".format(len(to_recrawl))
        except IndexError, exc:
            response['msg'] = str(exc)
            self._log_error("unexpected format of results form the 'what "
                            "accounts to recrawl' service ids query results", response)
            return

        response['error'] = False
        response['recrawls'] = to_recrawl
        self._write_result(response)


class PasteBinKeysToCountViewsController(DbConnectorHandler):

    def __init__(self, *args, **kwargs):
        self.service_name = "keys to measure for views"
        super(PasteBinKeysToCountViewsController, self).__init__(*args, **kwargs)

    @tornado.gen.coroutine
    def get(self):
        keys_future = models.doxes_for_view_measure(self.application.db)

        try:
            cursor = yield keys_future
            results = cursor.fetchall()
        except (psycopg2.Warning, psycopg2.Error) as exc:
            app_log.error("%(service_name)s:Database Error: %(msg)s", {
                "service_name": self.service_name,
                "msg": str(exc)
            })
            self.write({"error": True, "msg": str(exc)})
            return

        response = {"error": False, "msg": [r[0] for r in results]}
        self._write_result(response)


class RecordViewsForPasteController(DbConnectorHandler):

    def __init__(self, *args, **kwargs):
        self.service_name = "record views for paste"
        super(RecordViewsForPasteController, self).__init__(*args, **kwargs)

    @tornado.gen.coroutine
    def post(self):
        response = {"error": True, "msg": None}
        try:
            key = self.get_body_argument('key')
            views_count = int(self.get_body_argument('num_views'))
        except tornado.web.missingargumenterror, exc:
            response['msg'] = "missing post argument '{}'".format(exc.arg_name)
            self._log_error("error parsing submission", response)
            return
        except ValueError, exc:
            response['msg'] = "invalid integer value pasted for 'num_views'"
            self._log_error("error parsing submission", response)
            return

        paste_views_future = models.record_views_for_dox(self.application.db,
                                                         key, views_count)
        try:
            cursor = yield paste_views_future
            results = cursor.fetchall()
        except (psycopg2.Warning, psycopg2.Error) as exc:
            app_log.error("%(service_name)s:Database Error: %(msg)s", {
                "service_name": self.service_name,
                "msg": str(exc)
            })
            self.write({"error": True, "msg": str(exc)})
            return

        response = {"error": True, "msg": None}

        try:
            _ = results[0][0]
            response['msg'] = "Successfully recorded {} views for paste {}".format(views_count, key)
        except IndexError, exc:
            response['msg'] = str(exc)
            self._log_error("unexpected format of results form the 'record "
                            "views for paste' service", response)
            return

        response['error'] = False
        self._write_result(response)
