import logging
import pprint
import json
import HTMLParser
import subprocess
import requests
import bs4
import doxscrapers.utils


class BaseDoxScraper(object):


    def __init__(self, throttler_time, marshaller_host, source):
        self._marshaller = "http://{}/record".format(marshaller_host)
        self._source = source
        self._throttler = doxscrapers.utils.create_throttler(throttler_time)
        self._req_pause = 3
        self._logger = logging.getLogger("{}.log".format(source))


    def posts(self, limit=None, proxy=None):
        teasers = self.items_in_index(proxy=proxy)
        if not teasers:
            return None
        if not limit:
            limit = len(teasers)
        return [self.details_for_item(t) for t in teasers[:limit]]


    def throttle(self, wait_secs=None):
        self._throttler(wait_secs=wait_secs)


    def pause(self):
        self.throttle()


    def items_in_index(self, proxy=None):
        return NotImplemented


    def details_for_item(self, item, proxy=None):
        return NotImplemented


    def submit(self, item):
        payload = {
            "source": self._source,
            "text": item.to_json(),
        }

        item_id = item.key
        log_params = {
            "paste_id": item_id,
            "source": self._source
        }
        self._logger.info("Submitting paste %(paste_id)s from %(source)s", log_params)

        try:
            response_req = requests.post(self._marshaller, data=payload)
            response = response_req.json()
        except (requests.exceptions.HTTPError, requests.exceptions.Timeout), exc:
            log_params = {
                "paste_id": item_id,
                "error": str(exc)
            }
            self._logger.error("Received error submitting %(paste_id)s to "
                               "marshaller: %(error)s", log_params)
            return False
        except ValueError, exc:
            log_params = {
                "paste_id": item_id,
                "error": str(exc)
            }
            self._logger.error("Received invalid JSON when submitting %(paste_id)s to "
                               "the marshalling server: %(error)s", log_params)
            return False

        try:
            if response['error']:
                log_params = {
                    "paste_id": item_id,
                    "error": response['msg']
                }
                self._logger.error("Received error message when submitting "
                                   "%(paste_id)s to marshalling server: %(error)s",
                                   log_params)
                return False
        except KeyError:
            self._logger.error(u"Response fom the marshaller was not in the "
                               u"expected 'error', 'key', 'format'. Response: "
                               u"%(response)u", {
                                   u"response": pprint.pformat(response)
                               })
            return False

        self._logger.info("Successfuly submitted  %(paste_id)s: %(response)s", {
            "paste_id": item_id,
            "response": pprint.pformat(response)
        })
        return True


    def _request_url(self, url, proxy=None):
        try:
            return requests.get(url, proxies=proxy)
        except (requests.exceptions.HTTPError, requests.exceptions.Timeout,
                requests.exceptions.ConnectionError), exc:
            log_params = {
                u"url": url,
                u"error": str(exc)
            }
            self._logger.error(u"Error requesting url: %(url)s, "
                               u"received error: %(error)s", log_params)
            return None


    def _json_for_url(self, url, proxy=None):
        req_proxy = {"http": proxy} if proxy else None
        request_obj = self._request_url(url, proxy=req_proxy)
        if request_obj is None:
            return None

        try:
            return request_obj.json()
        except ValueError, exc:
            log_params = {
                u"url": url,
                u"error": str(exc)
            }
            self._logger.error(u"Error parsing JSON from url: %(url)s, "
                               u"received error: %(error)s", log_params)
            return None


    def _dom_for_url(self, url, proxy=None):
        request_text = self._request_url(url, proxy=proxy)
        if request_text is None:
            return None

        try:
            return bs4.BeautifulSoup(request_text.text, "html.parser")
        except HTMLParser.HTMLParseError, exc:
            log_params = {
                "url": url,
                "error": str(exc)
            }
            self._logger.error(u"Error parsing HTML when requesting %(url)s: "
                               u"%(error)s", log_params)
            return None


class TorDoxScraper(BaseDoxScraper):


    def pause(self):
        self.throttle(3600)


    def _request_url(self, url, *args):
        try:
            return subprocess.check_output(['torsocks', 'curl', '-s', url])
        except subprocess.CalledProcessError, exc:
            log_params = {
                "url": url,
                "error": str(exc)
            }
            self._logger.error(u"torsocks request to %(url)s failed with "
                               u"error: %(error)s", log_params)
            return None


    def _dom_for_url(self, url, *args):
        request_text = self._request_url(url)
        if request_text is None:
            return None

        try:
            return bs4.BeautifulSoup(request_text, "html.parser")
        except HTMLParser.HTMLParseError, exc:
            log_params = {
                "url": url,
                "error": str(exc)
            }
            self._logger.error(u"Error parsing HTML when requesting %(url)s: "
                               u"%(error)s", log_params)
            return None


    def details_for_item(self, item):
        details = dict(item)
        document = self._dom_for_url(item['full_url'])
        if document is None or document.find("textarea") is None:
            return None
        details['text'] = document.find("textarea").get_text()
        return details
